/* eslint-disable security/detect-object-injection */
'use strict';

/**
 * A very minimal API server framework.
 * Just a self-contained router and some request glue.
 */

require('./patches');
const { promises: fsPromises } = require('fs');
const path = require('path');
const querystring = require('querystring');
const common = require('./common');
const ContentNegotiation = require('./content-negotiation');
const Enum = require('./enum');
const { ResponseError, RouterNoPathError, RouterNoMethodError } = require('./errors');
const { extensionToMime } = require('./mime-helper');
const Router = require('./router');
const Template = require('./template');

// For logging.
const { fileScope } = require('@squeep/log-helper');
const _fileScope = fileScope(__filename);

const defaultOptions = {
  ignoreTrailingSlash: true,
  proxyPrefix: '',
  strictAccept: true,
  selfBaseUrl: '',
  staticMetadata: true,
  staticPath: undefined, // No reasonable default
  trustProxy: true,
  intrinsicHeadMethod: true,
  intrinsicHeadPersistBody: false,
  querystring,
};

const cookieSplitRE = /; */;

class Dingus {
  /**
   * @param {Object} logger object which implements logging methods
   * @param {Object} options
   * @param {Boolean} options.ignoreTrailingSlash requests for '/foo/' will match a '/foo' route
   * @param {string} options.proxyPrefix leading part of url path to strip
   * @param {Boolean} options.strictAccept whether to error on unsupported Accept type
   * @param {string} options.selfBaseUrl for constructing links
   * @param {Boolean} options.staticMetadata serve static headers with static files
   * @param {Boolean} options.trustProxy trust some header data to be provided by proxy
   * @param {Boolean} options.intrinsicHeadMethod handle HEAD requests automatically if not specified as a route method
   * @param {Boolean} options.intrinsicHeadPersistBody include un-sent body on ctx for automatic HEAD requests
   * @param {Object} options.querystring alternate qs parser to use
   */
  constructor(logger = console, options = {}) {
    common.setOptions(this, defaultOptions, options);

    this.router = new Router(options);

    if (!this.proxyPrefix) {
      this._stripPrefix = (p) => p;
    }

    this.responseTypes = [
      Enum.ContentType.TextHTML,
      Enum.ContentType.TextPlain,
      Enum.ContentType.ApplicationJson,
    ];

    this.logger = logger;
  }


  /**
   * Resolve relative and empty paths in url
   * @param {string} p path
   */
  _normalizePath(p) {
    const pathNorm = path.normalize(p); // This isn't perfectly correct, but it's easy...
    return this._stripPrefix(pathNorm);
  }


  /**
   * Remove a leading portion of url path
   * N.B. This method gets obliterated if there is no prefix defined at construction
   * @param {string} p path
   */
  _stripPrefix(p) {
    if (p.startsWith(this.proxyPrefix)) {
      return p.slice(this.proxyPrefix.length);
    }
    return p;
  }


  /**
   * Returns the path part, and querystring object, from a request url.
   * @param {string} url 
   */
  _splitUrl(url) {
    const [ p, qs ] = common.splitFirst(url, '?');
    return {
      pathPart: this._normalizePath(p),
      queryParams: this.querystring.parse(qs),
    };
  }


  /**
   * Insert a new path handler
   * @param {string} method 
   * @param {string} urlPath 
   * @param {fn} handler 
   */
  on(method, urlPath, handler, ...handlerArgs) {
    this.router.on(method, urlPath, handler, handlerArgs);
  }


  /**
   * Common header tagging for all requests.
   * Add our own identifier, and persist any external transit identifiers.
   * Sets requestId on ctx to a new uuid.
   * If X-Request-Id or X-Correlation-Id exist on incoming headers, sets them
   * on outgoing headers and sets on ctx.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  static tagContext(req, res, ctx) {
    const requestId = common.requestId();
    ctx.requestId = requestId;
    res.setHeader(Enum.Header.RequestId, requestId);
    [Enum.Header.XRequestId, Enum.Header.XCorrelationId].forEach((h) => {
      const v = req.getHeader(h);
      if (v) {
        ctx[h.replace(/-/g, '')] = v;
        res.setHeader(h, v);
      }
    });
    return requestId;
  }


  /**
   * 
   * @param {http.ClientRequest} req 
   */
  _getAddress(req) {
    // TODO: RFC7239 Forwarded support
    const address = (this.trustProxy && req?.getHeader(Enum.Header.XForwardedFor)) ||
      (this.trustProxy && req?.getHeader(Enum.Header.XRealIP)) ||
      (req?.connection?.remoteAddress) ||
      '';
    return address.split(/\s*,\s*/u)[0];
  }


  /**
   * 
   * @param {http.ClientRequest} req 
   */
  _getProtocol(req) {
    // TODO: RFC7239 Forwarded support
    const protocol = (this.trustProxy && req?.getHeader(Enum.Header.XForwardedProto)) ||
      ((req?.connection?.encrypted) ? 'https' : 'http');
    return protocol.split(/\s*,\s*/u)[0];
  }


  /**
   * Sets ctx.clientAddress and ctx.clientProtocol.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  clientAddressContext(req, res, ctx) {
    ctx.clientAddress = this._getAddress(req);
    ctx.clientProtocol = this._getProtocol(req);
  }


  /**
   * Sets ctx.cookie from Cookie header.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  static ingestCookie(req, res, ctx) {
    ctx.cookie = {};
    req.getHeader(Enum.Header.Cookie)?.split(cookieSplitRE).forEach((cookie) => {
      const [ name, value ] = common.splitFirst(cookie, '=', null).map((x) => {
        try {
          return decodeURIComponent(x.trim());
        } catch (e) {
          return x;
        }
      });
      if (name && !(name in ctx.cookie)) {
        const isQuoted = value?.startsWith('"') && value.endsWith('"');
        ctx.cookie[name] = isQuoted ? value.slice(1, -1) : value; // eslint-disable-line security/detect-object-injection
      }
    });
  }


  /**
   * Called before every request handler.
   * Sets tracking identifiers and client information on ctx.
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   */
  async preHandler(req, res, ctx) {
    this.constructor.tagContext(req, res, ctx);
    this.clientAddressContext(req, res, ctx);
    this.constructor.ingestCookie(req, res, ctx);
  }


  /**
   * Helper for collecting chunks as array of buffers.
   * @param {Buffer[]} chunks 
   * @param {string|Buffer} chunk 
   * @param {string} encoding 
   */
  static pushBufChunk(chunks, chunk, encoding = 'utf8') {
    if (chunk) {
      if (typeof chunk === 'string') {
        chunk = Buffer.from(chunk, encoding);
      }
      chunks.push(chunk);
    }
  }


  /**
   * Sets ctx.responseBody and calls handler upon res.end().
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   * @param {*} handler fn(req, res, ctx)
   */
  static setEndBodyHandler(req, res, ctx, handler) {
    const origWrite = res.write.bind(res);
    const origEnd = res.end.bind(res);
    const chunks = [];
    res.write = function (chunk, encoding, ...rest) {
      Dingus.pushBufChunk(chunks, chunk, encoding);
      return origWrite(chunk, encoding, ...rest);
    };
    res.end = function (data, encoding, ...rest) {
      Dingus.pushBufChunk(chunks, data, encoding);
      ctx.responseBody = Buffer.concat(chunks);
      handler(req, res, ctx);
      return origEnd(data, encoding, ...rest);
    };
  }


  /**
   * Intercept writes for head requests, do not send to client,
   * but send length, and make body available in context.
   * N.B. If persisted, ctx.responseBody will be a raw buffer, be aware when logging.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   * @param {Boolean} persistResponseBody
   */
  static setHeadHandler(req, res, ctx, persistResponseBody = false) {
    if (req.method === 'HEAD') {
      const origEnd = res.end.bind(res);
      const chunks = [];
      res.write = function (chunk, encoding) {
        Dingus.pushBufChunk(chunks, chunk, encoding);
        // No call to original res.write.
      };
      res.end = function (data, encoding, ...rest) {
        Dingus.pushBufChunk(chunks, data, encoding);
        const responseBody = Buffer.concat(chunks);
        res.setHeader(Enum.Header.ContentLength, Buffer.byteLength(responseBody));
        if (persistResponseBody) {
          ctx.responseBody = responseBody;
        }
        return origEnd(undefined, encoding, ...rest);
      };
    }
  }


  /**
   * Resolve the handler to invoke for a request.
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   * @returns {object}
   */
  _determineHandler(req, res, ctx) {
    const _scope = _fileScope('_determineHandler');

    const { pathPart, queryParams } = this._splitUrl(req.url);
    ctx.queryParams = queryParams;

    let handler, handlerArgs = [];
    try {
      ({ handler, handlerArgs } = this.router.lookup(req.method, pathPart, ctx));
    } catch (e) {
      if (e instanceof URIError) {
        handler = this.handlerBadRequest.bind(this);
      } else if (e instanceof RouterNoPathError) {
        handler = this.handlerNotFound.bind(this);
      } else if (e instanceof RouterNoMethodError) {
        if (this.intrinsicHeadMethod && req.method === 'HEAD') {
          ({ handler, handlerArgs } = this._determineHeadHandler(req, res, ctx, pathPart));
        } else {
          handler = this.handlerMethodNotAllowed.bind(this);
        }
      } else {
        this.logger.error(_scope, 'unexpected error', { error: e });
        handler = this.handlerInternalServerError.bind(this);
      }
    }
    return { handler, handlerArgs };
  }


  /**
   * For intrinsic HEAD requests, resolve the handler to invoke.
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   * @param {string} pathPart
   * @returns {object}
   */
  _determineHeadHandler(req, res, ctx, pathPart) {
    const _scope = _fileScope('_determineHeadHandler');
    let handler, handlerArgs = [];
    try {
      ({ handler, handlerArgs } = this.router.lookup('GET', pathPart, ctx));
      Dingus.setHeadHandler(req, res, ctx, this.intrinsicHeadPersistBody);
    } catch (e) {
      if (e instanceof RouterNoMethodError) {
        handler = this.handlerMethodNotAllowed.bind(this);
      } else {
        this.logger.error(_scope, 'unexpected error', { error: e });
        handler = this.handlerInternalServerError.bind(this);
      }
    }
    return { handler, handlerArgs };
  }


  /**
   * Dispatch the handler for a request
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   */
  async dispatch(req, res, ctx = {}) {
    const { handler, handlerArgs } = this._determineHandler(req, res, ctx);
    try {
      await this.preHandler(req, res, ctx);
      return await handler(req, res, ctx, ...handlerArgs);
    } catch (e) {
      ctx.error = e;
      this.sendErrorResponse(e, req, res, ctx);
    }
  }


  /**
   * Return normalized type, without any parameters.
   * @param {http.ClientRequest} req
   * @returns {string}
   */
  static getRequestContentType(req) {
    const contentType = req.getHeader(Enum.Header.ContentType);
    return (contentType || '').split(';')[0].trim().toLowerCase();
  }


  /**
   * Parse rawBody as contentType into ctx.parsedBody.
   * @param {string} contentType
   * @param {object} ctx
   * @param {string|buffer} rawBody
  */
  parseBody(contentType, ctx, rawBody) {
    const _scope = _fileScope('parseBody');

    switch (contentType) {
      case Enum.ContentType.ApplicationForm:
        ctx.parsedBody = this.querystring.parse(rawBody);
        break;

      case Enum.ContentType.ApplicationJson:
        try {
          ctx.parsedBody = JSON.parse(rawBody);
        } catch (e) {
          this.logger.debug(_scope, 'JSON parse failed', { requestId: ctx.requestId, error: e });
          throw new ResponseError(Enum.ErrorResponse.BadRequest, e.message);
        }
        break;

      default:
        this.logger.debug(_scope, 'unhandled content-type', { requestId: ctx.requestId, contentType });
        throw new ResponseError(Enum.ErrorResponse.UnsupportedMediaType);
    }
  }


  /**
   * Return all body data from a request.
   * @param {http.ClientRequest} req
   * @param {Number=} maximumBodySize
   * @param {Boolean=} toString
   */
  async bodyData(req, maximumBodySize, toString = true) {
    const _scope = _fileScope('bodyData');
    return new Promise((resolve, reject) => {
      const body = [];
      let length = 0;
      req.on('data', (chunk) => {
        body.push(chunk);
        length += Buffer.byteLength(chunk);
        if (maximumBodySize && length > maximumBodySize) {
          this.logger.debug(_scope, 'body data exceeded limit', { length, maximumBodySize });
          reject(new ResponseError(Enum.ErrorResponse.RequestEntityTooLarge));
        }
      });
      req.on('end', () => {
        const bodyBuffer = Buffer.concat(body);
        resolve(toString ? bodyBuffer.toString() : bodyBuffer);
      });
      req.on('error', (e) => {
        this.logger.error(_scope, 'failed', { error: e });
        reject(e);
      });
    });
  }


  /**
   * Read and parse request body data.
   * Sets ctx.parsedBody, and optionally ctx.rawBody.
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   * @param {object}
   * @param {Boolean} .parseEmptyBody
   * @param {Boolean} .persistRawBody
   */
  async ingestBody(req, res, ctx, { parseEmptyBody = true, persistRawBody = false, maximumBodySize } = {}) {
    const rawBody = await this.bodyData(req, maximumBodySize);
    if (persistRawBody) {
      ctx.rawBody = rawBody;
    }
    if (rawBody || parseEmptyBody) {
      const contentType = Dingus.getRequestContentType(req);
      this.parseBody(contentType, ctx, rawBody);
    }
  }


  /**
   * Return the best matching response type.
   * @param {string[]} responseTypes 
   * @param {http.ClientRequest} req 
   */
  static getResponseContentType(responseTypes, req) {
    const acceptHeader = req.getHeader(Enum.Header.Accept);
    return ContentNegotiation.accept(responseTypes, acceptHeader);
  }


  /**
   * Returns a list of the most-preferred content encodings for the response.
   * @param {string[]} responseEncodings
   * @param {http.ClientRequest} req 
   */
  static getResponseEncoding(responseEncodings, req) {
    const acceptEncodingHeader = req.getHeader(Enum.Header.AcceptEncoding);
    return ContentNegotiation.preferred(responseEncodings, acceptEncodingHeader);
  }


  /**
   * Set the best content type for the response.
   * Sets ctx.responseType, and Content-Type header.
   * @param {string[]} responseTypes default first
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  setResponseType(responseTypes, req, res, ctx) {
    const _scope = _fileScope('setResponseType');
    ctx.responseType = Dingus.getResponseContentType(responseTypes, req);
    if (!ctx.responseType) {
      if (this.strictAccept) {
        this.logger.debug(_scope, 'unhandled strict accept', { requestId: req.requestId });
        throw new ResponseError(Enum.ErrorResponse.NotAcceptable);
      } else {
        ctx.responseType = responseTypes[0];
      }
    }
    res.setHeader(Enum.Header.ContentType, ctx.responseType);
  }


  /**
   * Inserts an encoding into Content-Encoding header.
   * @param {http.ServerResponse} res
   * @param {string} encoding
   */
  static addEncodingHeader(res, encoding) {
    const existingEncodings = res.getHeader(Enum.Header.ContentEncoding);
    if (existingEncodings) {
      encoding = `${encoding}, ${existingEncodings}`;
    }
    res.setHeader(Enum.Header.ContentEncoding, encoding);
  }


  /**
   * Attempt to fetch both data and metadata for a file.
   * @param {string} filePath 
   */
  async _readFileInfo(filePath) {
    const _scope = _fileScope('_readFileInfo');
    let result;
    try {
      // eslint-disable-next-line security/detect-non-literal-fs-filename
      const stat = fsPromises.stat(filePath);
      // eslint-disable-next-line security/detect-non-literal-fs-filename
      const data = fsPromises.readFile(filePath);
      result = await Promise.all([stat, data]);
    } catch (e) {
      if (['ENOENT', 'EACCES', 'EISDIR', 'ENAMETOOLONG', 'EINVAL'].includes(e.code)) {
        return [null, null];
      }
      this.logger.error(_scope, 'fs error', { error: e, filePath });
      throw e;
    }
    return result;
  }


  /**
   * Potentially add additional headers from static file meta-file.
   * @param {http.ServerResponse} res
   * @param {string} directory
   * @param {string} fileName - already normalized and filtered
   */
  async _serveFileMetaHeaders(res, directory, fileName) {
    const _scope = _fileScope('_serveFileMetaHeaders');

    let added = false;
    const metaPrefix = '.';
    const metaSuffix = '.meta';
    const metaFileName = `${metaPrefix}${fileName}${metaSuffix}`;
    const metaFilePath = path.join(directory, metaFileName);

    const [stat, data] = await this._readFileInfo(metaFilePath);
    if (!stat) {
      return added;
    }

    const lineBreakRE = /\r\n|\n|\r/;
    const lines = data.toString().split(lineBreakRE);
    common.unfoldHeaderLines(lines);

    const headerParseRE = /^(?<name>[^:]+): +(?<value>.*)$/;
    lines.forEach((line) => {
      if (line) {
        const result = headerParseRE.exec(line);
        const { groups: header } = result;
        res.setHeader(header.name, header.value);
        added = true;
      }
    });
    return added;
  }


  /**
   * Serve a file from a directory, with rudimentary cache awareness.
   * This will also serve pre-encoded variations if available and requested.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   * @param {string} directory 
   * @param {string} fileName 
   */
  async serveFile(req, res, ctx, directory, fileName) {
    const _scope = _fileScope('serveFile');
    this.logger.debug(_scope, 'called', { req, ctx });

    // Require a directory field.
    if (!directory) {
      this.logger.debug(_scope, 'rejected unset directory', { fileName });
      return this.handlerNotFound(req, res, ctx);
    }

    // Normalize the supplied path, as encoded path-navigation may have been (maliciously) present.
    fileName = path.normalize(fileName);

    // We will not deal with any subdirs, nor any dot-files.
    // (Note that we could not deal with subdirs even if we wanted, due to simple router matching scheme.)
    if (fileName.indexOf(path.sep) >= 0
    ||  fileName.startsWith('.')) {
      this.logger.debug(_scope, 'rejected filename', { fileName });
      return this.handlerNotFound(req, res, ctx);
    }

    const filePath = path.join(directory, fileName);

    // File must exist, before any alternate static encodings will be considered.
    let [stat, data] = await this._readFileInfo(filePath);
    if (!stat) {
      return this.handlerNotFound(req, res, ctx);
    }

    // If encodings were requested, check for static versions to serve.
    // Update stat and data if matching version is found.
    ctx.availableEncodings = Dingus.getResponseEncoding(Object.values(Enum.EncodingType), req);
    if (ctx.availableEncodings.length === 0) {
      // Identity encoding was specifically denied, and nothing else available.
      this.logger.debug(_scope, 'no suitable encodings', { ctx });
      return this.handlerMethodNotAllowed(req, res, ctx);
    }
    for (const encoding of ctx.availableEncodings) {
      if (encoding === Enum.EncodingType.Identity) {
        break;
      }
      const suffix = Enum.EncodingTypeSuffix[encoding];
      if (!suffix) {
        this.logger.error(_scope, 'supported encoding missing mapped suffix', { ctx, encoding });
        continue;
      }
      const encodedFilePath = `${filePath}${suffix}`;
      const [ encodedStat, encodedData ] = await this._readFileInfo(encodedFilePath);
      if (encodedStat) {
        ([ stat, data ] = [ encodedStat, encodedData ]);
        ctx.selectedEncoding = encoding;
        Dingus.addEncodingHeader(res, encoding);
        res.setHeader(Enum.Header.Vary, Enum.Header.AcceptEncoding);
        this.logger.debug(_scope, 'serving encoded version', { ctx, encodedFilePath });
      }
      break;
    }

    const lastModifiedDate = new Date(stat.mtimeMs);
    res.setHeader(Enum.Header.LastModified, lastModifiedDate.toGMTString());

    const eTag = common.generateETag(filePath, stat, data);
    res.setHeader(Enum.Header.ETag, eTag);

    if (common.isClientCached(req, stat.mtimeMs, eTag)) {
      this.logger.debug(_scope, 'client cached file', { filePath });
      res.statusCode = 304; // Not Modified
      res.end();
      return;
    }

    // Set the type based on extension of un-encoded filename.
    const ext = path.extname(filePath).slice(1); // Drop the dot
    const contentType = extensionToMime(ext);
    res.setHeader(Enum.Header.ContentType, contentType);

    // We presume static files are relatively cacheable.
    res.setHeader(Enum.Header.CacheControl, 'public');

    if (this.staticMetadata) {
      ctx.metaHeaders = await this._serveFileMetaHeaders(res, directory, fileName);
    }

    this.logger.debug(_scope, 'serving file', { filePath, contentType });
    res.end(data);
  }


  /**
   * Return a content-type appropriate rendering of an errorResponse object.
   * @param {string} type content-type of response
   * @param {object} err either an Error object, or an error response
   * @param {number} err.statusCode
   * @param {string} err.errorMessage
   * @param {string|string[]} err.details
   */
  // eslint-disable-next-line class-methods-use-this
  renderError(contentType, err) {
    switch (contentType) {
      case Enum.ContentType.ApplicationJson:
        return JSON.stringify(err);

      case Enum.ContentType.TextHTML:
        return Template.errorHTML(err);

      case Enum.ContentType.TextPlain:
      default:
        return [err.errorMessage, err.details].join('\r\n');
    }
  }


  /**
   * Send an error response.  Terminal.
   * Logs any non-error-response errors as such.
   * @param {object} err either an Error object, or an error response
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  sendErrorResponse(err, req, res, ctx) {
    const _scope = _fileScope('sendErrorResponse');
    let body;

    // Default to a content type if one is not yet present
    if (!res.hasHeader(Enum.Header.ContentType)) {
      res.setHeader(Enum.Header.ContentType, Enum.ContentType.TextPlain);
    }

    if (err?.statusCode) {
      res.statusCode = err.statusCode;
      body = this.renderError(res.getHeader(Enum.Header.ContentType), err);
      this.logger.debug(_scope, 'handler error', { err, req, res, ctx });
    } else {
      res.statusCode = 500;
      body = this.renderError(res.getHeader(Enum.Header.ContentType), Enum.ErrorResponse.InternalServerError);
      this.logger.error(_scope, 'handler exception', { err, req, res, ctx });
    }
    res.end(body);
  }


  /**
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   * @param {String} file - override ctx.params.file
   */
  async handlerGetStaticFile(req, res, ctx, file) {
    Dingus.setHeadHandler(req, res, ctx);

    // Set a default response type to handle any errors; will be re-set to serve actual static content type.
    this.setResponseType(this.responseTypes, req, res, ctx);

    await this.serveFile(req, res, ctx, this.staticPath, file || ctx.params.file);
  }


  /**
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {Object} ctx
   * @param {String} newPath
   * @param {Number} statusCode
  */
  async handlerRedirect(req, res, ctx, newPath, statusCode = 307) {
    this.setResponseType(this.responseTypes, req, res, ctx);
    res.setHeader(Enum.Header.Location, newPath);
    res.statusCode = statusCode;
    res.end();
  }


  /**
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   */
  async handlerMethodNotAllowed(req, res, ctx) {
    this.setResponseType(this.responseTypes, req, res, ctx);
    throw new ResponseError(Enum.ErrorResponse.MethodNotAllowed);
  }


  /**
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   */
  async handlerNotFound(req, res, ctx) {
    this.setResponseType(this.responseTypes, req, res, ctx);
    throw new ResponseError(Enum.ErrorResponse.NotFound);
  }


  /**
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   */
  async handlerBadRequest(req, res, ctx) {
    this.setResponseType(this.responseTypes, req, res, ctx);
    throw new ResponseError(Enum.ErrorResponse.BadRequest);
  }


  /**
   * @param {http.ClientRequest} req
   * @param {http.ServerResponse} res
   * @param {object} ctx
   */
  async handlerInternalServerError(req, res, ctx) {
    this.setResponseType(this.responseTypes, req, res, ctx);
    throw new ResponseError(Enum.ErrorResponse.InternalServerError);
  }

}

module.exports = Dingus;
